print("import: 'hgdemandimport'")
import hgdemandimport

print("import: 'hgext'")
import hgext

print("import: 'hgext.convert'")
import hgext.convert

print("import: 'hgext.fsmonitor'")
import hgext.fsmonitor

print("import: 'hgext.fsmonitor.pywatchman'")
import hgext.fsmonitor.pywatchman

print("import: 'hgext.fsmonitor.pywatchman.bser'")
import hgext.fsmonitor.pywatchman.bser

print("import: 'hgext.highlight'")
import hgext.highlight

print("import: 'hgext.largefiles'")
import hgext.largefiles

print("import: 'hgext.zeroconf'")
import hgext.zeroconf

print("import: 'hgext3rd'")
import hgext3rd

print("import: 'mercurial'")
import mercurial

print("import: 'mercurial.cext'")
import mercurial.cext

print("import: 'mercurial.cffi'")
import mercurial.cffi

print("import: 'mercurial.hgweb'")
import mercurial.hgweb

print("import: 'mercurial.httpclient'")
import mercurial.httpclient

print("import: 'mercurial.pure'")
import mercurial.pure

